#include <string>
#include <math.h>
#include <fstream>
#include <boost/iostreams/filter/zlib.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>
#include <boost/iostreams/copy.hpp>
#include <vector>
#include <set>
#include <sstream>
#include <iostream>
#include <map>
#include <chrono>
#include <boost/thread/mutex.hpp>

boost::mutex mutex;

struct Vec4{
    public:
        double x, y, z, w;
        Vec4(double a, double b, double c, double d){
            x=a;
            y=b;
            z=c;
            w=d;
        }
};

struct tripletInt{
    int x, y, z;
    tripletInt(){}
    tripletInt(int a, int b, int c){
        x=a; y=b; z=c;
    }
    bool operator==(const tripletInt &other) const
    { return (x == other.x
                && y == other.y &&
                z == other.z);
    }

    bool operator<(const tripletInt& other) const
    {
        return x<other.x ||
            (x == other.x && y<other.y) ||
            (x == other.x && y==other.y && z<other.z);
    }
};


std::string gasTypesByCode[14] = {
        "ethanol",
		"methane",
		"hydrogen",
		"propanol",
		"chlorine",
		"flurorine",
		"acetone",
		"neon",
		"helium",
		"hot_air",
		"biogas",
		"carbon dioxide",
		"carbon monoxide",
        "smoke"
    };

//gas and wind simulation variables
std::vector<Vec4> activeFilaments;
double total_moles_in_filament, num_moles_all_gases_in_cm3;
std::vector<uint8_t> Env;
std::vector<double> C;  //3D Gas concentration
std::vector<double> U;  //3D Wind U
std::vector<double> V;  //3D Wind V
std::vector<double> W;  //3D Wind W
double env_min_x, env_min_y, env_min_z;
double env_max_x, env_max_y, env_max_z;
int environment_cells_x, environment_cells_y, environment_cells_z;
double environment_cell_size;
std::string gas_type;




//paths and auxiliary variables for parsing and printing
std::string gasSimulationsPath;
int last_wind_idx=-1;

int indexFrom3D(int x, int y, int z){
    return x + y*environment_cells_x + z*environment_cells_x*environment_cells_y;
}


void decompress(std::string filename, std::stringstream& decompressed){
    std::ifstream infile(filename, std::ios_base::binary);
    boost::iostreams::filtering_streambuf<boost::iostreams::input> inbuf;
    inbuf.push(boost::iostreams::zlib_decompressor());
    inbuf.push(infile);

    boost::iostreams::copy(inbuf,decompressed);
    infile.close();
}

void readOccupancyFileAndHeaders(std::string occupancyFile, std::string& filename){

    std::stringstream decompressed;
    decompress(filename, decompressed);

    decompressed.seekg(sizeof(int));
    double bufferD [5];
    decompressed.read((char*) &env_min_x, sizeof(double));
    decompressed.read((char*) &env_min_y, sizeof(double));
    decompressed.read((char*) &env_min_z, sizeof(double));

    decompressed.read((char*) &env_max_x, sizeof(double));
    decompressed.read((char*) &env_max_y, sizeof(double));
    decompressed.read((char*) &env_max_z, sizeof(double));

    decompressed.read((char*) &environment_cells_x, sizeof(int));
    decompressed.read((char*) &environment_cells_y, sizeof(int));
    decompressed.read((char*) &environment_cells_z, sizeof(int));

    decompressed.read((char*) &environment_cell_size, sizeof(double));
    
    decompressed.read((char*) &bufferD, 5*sizeof(double));
    int gt;
    decompressed.read((char*) &gt, sizeof(int));
    gas_type=gasTypesByCode[gt];

    decompressed.read((char*) &total_moles_in_filament, sizeof(double));
    decompressed.read((char*) &num_moles_all_gases_in_cm3, sizeof(double));

    Env.resize(environment_cells_x * environment_cells_y * environment_cells_z);

    if(FILE *file = fopen(filename.c_str(), "r")){
        fclose(file);
    }
    else{
        std::cout<<"File "<<filename<< " does not exist.\n\n";
        return;
    }

	//open file
	std::ifstream infile(occupancyFile.c_str());
	std::string line;

    //discard the header
    std::getline(infile, line);
    std::getline(infile, line);
    std::getline(infile, line);
    std::getline(infile, line);

    int x_idx = 0;
	int y_idx = 0;
	int z_idx = 0;

	while ( std::getline(infile, line) )
	{
		std::stringstream ss(line);

		if (line == ";")
		{
			//New Z-layer
			z_idx++;
			x_idx = 0;
			y_idx = 0;
		}
		else
		{   //New line with constant x_idx and all the y_idx values
			while (!ss.fail())
			{
				double f;
				ss >> f;		//get one double value
				if (!ss.fail())
				{
					Env[indexFrom3D(x_idx, y_idx, z_idx)] = f;
					y_idx++;
				}
			}

			//Line has ended
			x_idx++;
			y_idx = 0;
		}
	}
    infile.close();

    C.resize(environment_cells_x * environment_cells_y * environment_cells_z);
    U.resize(environment_cells_x * environment_cells_y * environment_cells_z);
    V.resize(environment_cells_x * environment_cells_y * environment_cells_z);
    W.resize(environment_cells_x * environment_cells_y * environment_cells_z);
}


void load_wind_file(int wind_index){
    if(wind_index==last_wind_idx)
        return;
    last_wind_idx=wind_index;

    std::ifstream infile(boost::str( boost::format("%s/wind/wind_iteration_%i") % gasSimulationsPath.c_str() % wind_index), std::ios_base::binary);
    infile.read((char*) U.data(), sizeof(double)* U.size());
    infile.read((char*) V.data(), sizeof(double)* V.size());
    infile.read((char*) W.data(), sizeof(double)* W.size());
    infile.close();
}


void load_gas_file(std::stringstream& decompressed){
    decompressed.seekg(14*sizeof(double) + 5*sizeof(int));
    int wind_index;
    decompressed.read((char*) &wind_index, sizeof(int));

    activeFilaments.clear();
    int filament_index;
    double x, y, z, stdDev;
    while(decompressed.peek()!=EOF){

        decompressed.read((char*) &filament_index, sizeof(int));
        decompressed.read((char*) &x, sizeof(double));
        decompressed.read((char*) &y, sizeof(double));
        decompressed.read((char*) &z, sizeof(double));
        decompressed.read((char*) &stdDev, sizeof(double));

        activeFilaments.emplace_back(x, y, z, stdDev);
    }

    load_wind_file(wind_index);
}





int check_pose_with_environment(double pose_x, double pose_y, double pose_z)
{
	//1.1 Check that pose is within the boundingbox environment
	if (pose_x<env_min_x || pose_x>env_max_x || pose_y<env_min_y || pose_y>env_max_y || pose_z<env_min_z || pose_z>env_max_z)
		return 1;

	//Get 3D cell of the point
	int x_idx = (pose_x-env_min_x)/environment_cell_size;
	int y_idx = (pose_y-env_min_y)/environment_cell_size;
	int z_idx = (pose_z-env_min_z)/environment_cell_size;

	if (x_idx >= environment_cells_x || y_idx >= environment_cells_y || z_idx >= environment_cells_z)
		return 1;

	//1.2. Return cell occupancy (0=free, 1=obstacle, 2=outlet)
	return Env[indexFrom3D(x_idx,y_idx,z_idx)];
}

bool check_environment_for_obstacle(double start_x, double start_y, double start_z,
													   double   end_x, double   end_y, double end_z)
{

	// Check whether one of the points is outside the valid environment or is not free
	if(check_pose_with_environment(start_x, start_y, start_z) != 0)   { return false; }
	if(check_pose_with_environment(  end_x, end_y  ,   end_z) != 0)   { return false; }


	// Calculate normal displacement vector
	double vector_x = end_x - start_x;
	double vector_y = end_y - start_y;
	double vector_z = end_z - start_z;
	double distance = sqrt(vector_x*vector_x + vector_y*vector_y + vector_z*vector_z);
	vector_x = vector_x/distance;
	vector_y = vector_y/distance;
	vector_z = vector_z/distance;


	// Traverse path
	int steps = ceil( distance / environment_cell_size );	// Make sure no two iteration steps are separated more than 1 cell
	double increment = distance/steps;

	for(int i=1; i<steps-1; i++)
	{
		// Determine point in space to evaluate
		double pose_x = start_x + vector_x*increment*i;
		double pose_y = start_y + vector_y*increment*i;
		double pose_z = start_z + vector_z*increment*i;


		// Determine cell to evaluate (some cells might get evaluated twice due to the current code
		int x_idx = (pose_x-env_min_x)/environment_cell_size ;
		int y_idx = (pose_y-env_min_y)/environment_cell_size ;
		int z_idx = (pose_z-env_min_z)/environment_cell_size ;


		// Check if the cell is occupied
		if(Env[indexFrom3D(x_idx,y_idx,z_idx)] != 0) { return false; }
	}

	// Direct line of sight confirmed!
	return true;
}





double concentration_from_filament(float x, float y, float z, Vec4 filament){
    //calculate how much gas concentration does one filament contribute to the queried location
    double sigma = filament.w;
    double distance_cm = 100 * sqrt( pow(x-filament.x,2) + pow(y-filament.y,2) + pow(z-filament.z,2) );

    double num_moles_target_cm3 = (total_moles_in_filament /
        (sqrt(8*pow(M_PI,3)) * pow(sigma,3) )) * exp( -pow(distance_cm,2)/(2*pow(sigma,2)) );

    double ppm = num_moles_target_cm3/num_moles_all_gases_in_cm3 * 1000000; //parts of target gas per million

    return ppm;
}

void calculateAllConcentrations(){
    for(auto it = activeFilaments.begin(); it!=activeFilaments.end(); it++){
        Vec4 fil = *it;
        double limitDistance = fil.w*3/100;
        int min_x = std::max(0, (int)((fil.x-limitDistance - env_min_x)/environment_cell_size) );
        int min_y = std::max(0, (int)((fil.y-limitDistance - env_min_y)/environment_cell_size) );
        int min_z = std::max(0, (int)((fil.z-limitDistance - env_min_z)/environment_cell_size) );


        int max_x = std::min(environment_cells_x, (int)((fil.x + limitDistance - env_min_x)/environment_cell_size) );
        int max_y = std::min(environment_cells_y, (int)((fil.y + limitDistance - env_min_y)/environment_cell_size) );
        int max_z = std::min(environment_cells_z, (int)((fil.z + limitDistance - env_min_z)/environment_cell_size) );


        #pragma omp parallel for collapse(3)
        for(int i=min_x; i<max_x; i++){
            for(int j=min_y; j<max_y; j++){
                for(int k=min_z; k<max_z; k++){
                    double pos_x = env_min_x + (i+0.5)*environment_cell_size; 
                    double pos_y = env_min_y + (j+0.5)*environment_cell_size;
                    double pos_z = env_min_z + (k+0.5)*environment_cell_size;
                    
                    if(check_environment_for_obstacle(fil.x, fil.y, fil.z, pos_x, pos_y, pos_z)){
                        float c = concentration_from_filament(pos_x, pos_y, pos_z,
                                                fil);
                        mutex.lock();
                            C[indexFrom3D(i,j,k)] += c;
                        mutex.unlock();
                    }
                }
            }
        }

    }
}





void writeFile(std::string& filename){
    FILE* file = fopen(filename.c_str(),"w+");
    fprintf(file, "x_idx,y_idx,z_idx,Concentration(ppm),U(m/s),V(m/s),W(m/s)\n\n");
    for (int i=0; i<environment_cells_x; i++){
        for (int j=0; j<environment_cells_y; j++){
            for (int k=0; k<environment_cells_z; k++){
                if(Env[indexFrom3D(i,j,k)]!=1)
                    fprintf(file, "%d,%d,%d,%f,%f,%f,%f\n", i, j, k, C[indexFrom3D(i,j,k)], U[indexFrom3D(i,j,k)], V[indexFrom3D(i,j,k)], W[indexFrom3D(i,j,k)]);
            }
        }
    }
    fclose(file);
}


int main(int argc, char const *argv[])
{
    if(argc!=4){
        std::cout<<"Wrong input format.\nDo: writeConcentrations (occupancyFile) (simulationFolder) (outputFolder) ";
        return -1;
    }
    std::string occupancyFile = argv[1]; 
    gasSimulationsPath = argv[2];
    std::string outputPath = argv[3];
    
    std::string filename = gasSimulationsPath+"/iteration_0";
    int i=0;
    readOccupancyFileAndHeaders(occupancyFile, filename);
    
    auto start = std::chrono::system_clock::now();
    while(FILE *file = fopen(filename.c_str(), "r")) {
        fclose(file);
        
        std::stringstream decompressed;
        decompress(filename, decompressed);
        load_gas_file(decompressed);


        calculateAllConcentrations();
        std::string outfile = boost::str(boost::format("%s/iteration_%d.csv") % outputPath % i);
        writeFile(outfile);
        
        i++;
        filename = boost::str(boost::format("%s/iteration_%d") % gasSimulationsPath % i);
        if (i==200){
            std::cout << (std::chrono::system_clock::now() - start).count()/((double) pow(10,9)) <<"\n";
        }
    }

   
    return 0;
}
